(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/flatten.sml,v
   Comment: Flattener for Absyn models (Absyn.sml -> flat.sml)
   Release: scope-0-16_final 

     Copyright (C) 2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   Besides some typical C output tools, it contains a function to generate
   eexternal functions referenced in source code. The functions generated 
   are empty return void or return true, depending on type.

   FIXME: wszystkie odwolania do BDD pewnie mozna wywalic
   
   flatten.sml,v 1.26 2003/07/29 12:46:45 wasowski Exp
*)

(* expand_initial : tran -> tran
   remove_entry : tran -> tran 
   remove_exit : tran -> tran list BLOW UP
   flatten_guard : tran -> tran
   flatten : Absyn.model -> flat model *)

(* type of flat models *)

structure Flatten = struct

open Absyn Err Haskellish Say 

structure F = Flat

open WuddySys ListPartial

infix 7 AND
infix 6 OR
infix 4 IMP

		

(* Set based optimizer ------------------------------------------------- *)

(* This is an optimizer of DNF clauses (i.e. long conjunctions of terms).
   It actually returns optimal clause, i.e. the one with shortest
   representation. It has additionally preference of positive terms against
   negative terms which gives slightly better results in flattening 

   See another (worse) optimizer below. The other one carries all analysis
   with BDDs while this one only uses BDDs to reason about hierarchical
   model. (Also I start to think that similar machinery can be proposed for
   optimizing hierarchical models).
   
*)
	
(* FIXME Implementation of mconf can certainly be improved if needed *)
(* FIXME If works out should placed in separate structure            *)

datatype mconf = Sat of (Set.set * Set.set) env | NonSat 

(* Invariant: states are never deleted, they are moved btw sets *)
(*	      you can only move states from pos to neg sets     *)
(*            empty pos set means contradiction (FALSE)         *)
(*            empty neg set means tautology (TRUE)              *) 

fun conjoinVar (a :F.anatomy) (s :stateid,NonSat   ) = NonSat
  | conjoinVar (a :F.anatomy) (s :stateid,Sat mconf) =
	let val mid   = SE.lookup a s
	    val (p,n) = SE.lookup mconf mid
			handle e => intErr e 200307073
	    val ss    = Set.singleton s
	in 
		if (Set.member (p,s)) then Sat (SE.bind1 mconf 
				(mid,(ss,Set.delete(Set.union(p,n),s))))
		else NonSat
	end 
	handle NotFound => Sat mconf 
				 (* s belongs to single state machine,
				    optimized out -- always true *)

fun conjoinNVar (a :F.anatomy) (s :stateid,NonSat   ) = NonSat 
  | conjoinNVar (a :F.anatomy) (s :stateid,Sat mconf) =
	let val mid   = SE.lookup a s
	    val (p,n) = SE.lookup mconf mid
			handle e => intErr e 200307072
	in if (Set.member (n,s)) then Sat mconf
	   else let val p1 = Set.delete (p,s)
			     handle e => intErr e 200307071
		in if Set.isEmpty p1 then NonSat else 
			Sat (SE.bind1 mconf (mid,(p1, Set.add(n,s))))
		end
	end handle NotFound => NonSat (* s belongs to a single state SEM
					 optimized out -- never false *)

(* build a tautology *)

fun build_taut (a :F.anatomy) :mconf =
	let val smlist = SE.toList a
	    fun add ((s,m),env) =
		SE.bind1 env (m,Set.add (SE.lookup env m,s))
		handle NotFound => SE.bind1 env (m, Set.singleton s)
	    val posconf = foldl add SE.empty smlist
	in Sat (SE.map (fn p=>(p,Set.empty)) posconf) end

(* this is faster but needs hierarchical model     *)
(* and cannot be used for nonisomorphic flattening *)

fun build_tauth (h :'aor hierarchy) :mconf =
	let fun fo (s :'aor orstate,e) =
		let val kids = map nameand (childrenor s)
		in  
		    if length kids > 1 then
			SE.bind1 e (nameor s,(Set.fromList kids,Set.empty))
		    else e
		end
	in Sat (hfoldand fo #2 SE.empty (gettree h)) end

(* translate mconf to posneg (semantics to syntax) *)

fun mconf2posneg NonSat :posneg option = NONE
  | mconf2posneg (Sat mconf) = 
	let fun mksem ((p,n),(pos,neg)) = 
		if Set.isEmpty n then (pos,neg)
		else if Set.numItems p = 1 then 
			(hd (Set.listItems p)::pos,neg)
		else (pos, Set.listItems n @ neg) (* Set.numItems p > 1 *) 
	    val sems = map #2 (SE.toList mconf)
	in SOME (foldl mksem ([],[]) sems) end
	 
(* an optimizer itself *)		

fun optimize1_cond (sat,a :F.anatomy) (pos,neg) :posneg option =
	let val p  = foldl (conjoinVar  a) sat pos
	    val pn = foldl (conjoinNVar a) p   neg
	in mconf2posneg pn end handle e => intErr e 200307074

(* Optimize a bunch of guards                                *)
(* FIXME: still should attempt to marry DNF clauses into one *)

fun optimize1_conds (a :F.anatomy,sat :mconf) (conds :posneg list) 
								:posneg list =
	mapPartial (optimize1_cond (sat,a)) conds

(* End of mconf based (set-based) optimizer                                *)
(* ----------------------------------------------------------------------- *)

(* -------------------------------------------------------------------------- *)
(* BDD-based optimizer

   flat guard optimizer. This optimizer is nice because it actually
   guarantees  generation of minimal guard from a given guard. It also
   eliminates unsound and tautology guards. Still it does not join guards
   if possible (from two or more DNF clauses into one) 
   
   PROBLEM: the minimality of guard is weak here. It is minimal in the sense
   that nothing can be removed from it, but not in the sense that it is the
   shortest equivalent clause. See another set-based implementation for 
   stronger optimality *)

local 
  
  fun add_cond (((p,n),cond), old as (gamma,psi_f,psi_h,(pos,neg)))  =
	if always (gamma IMP cond) then SOME old
	else if always (psi_f IMP cond) then SOME old
	else if never  (psi_h AND cond) then NONE
	else if never  (psi_f AND cond) then NONE
	else SOME (gamma AND cond, psi_f AND cond, psi_h AND cond, (p@pos,n@neg)) 
in
  
  (* incremental optimizers *)

  fun add_pos (s :stateid, optimized_part : bdd*bdd*bdd*posneg) 
				: (bdd * bdd * bdd * posneg) option =
	add_cond ((([s],[]),bddVar s), optimized_part)	

  fun add_neg (s :stateid, optimized_part : bdd*bdd*bdd*posneg)
				: (bdd * bdd * bdd * posneg) option =
	add_cond ((([],[s]),bddNVar s),optimized_part)

  (* whole guard optimizer *)

  fun optimize_cond (phi_f,phi_h) ((pos,neg) :posneg) =
	let val res = foldlPartial add_pos (W.TRUE,phi_f,phi_h,([],[])) pos
	in foldlPartial add_neg (valOf res) neg end
	handle Option.Option => NONE

  (* FIXME: still should attempt to marry DNF clauses into one *)

  fun optimize_conds (phi_f,phi_h) (conds :posneg list) :posneg list = 
	map #4 (mapPartial (optimize_cond (phi_f,phi_h)) conds)
	
end

(* End of BDD based optimizer ----------------------------------------------- *)
(* -------------------------------------------------------------------------- *)



(* -------------------------------------------------------------------------- *)
(* Actual flattening algorithm ---------------------------------------------- *)

(* compute flat system's structure (groups of sets of states) *)

fun fl_hierarchy (h :'aor hierarchy) :F.anatomy * F.machineid list = 
	let fun fo (s :'aor orstate, (a,ms)) :F.anatomy * F.machineid list =  
		let val no    = nameor s
		    val assoc = map (fn s=> (nameand s,no)) (childrenor s)
		in if length assoc = 1 then (a,ms)
		   else (SE.bind a assoc,no::ms)
		end
	in hfoldand fo #2 (SE.empty,[]) (gettree h) end

(* flatten transition's condition, nonoptimized result    *)

fun posneg_closure (h :'aor hierarchy) ((pos,neg) :posneg) :posneg list =
	let val pos1 = List.concat (map (ancest2hst h) pos) 
	    fun mkneg1 (ns :stateid, pns :posneg list) :posneg list = 
		let val pns1 = map (fn (p,n)=>(ancest2h h ns @ p,ns::n)) pns
		    val pns2 = mkneg1 (parenth h (parenth h ns),pns) 
				handle StateNotFound => []
		in pns2 @ pns1 end
	in foldl mkneg1 [(pos1,[])] neg end

(* manual test procedures for closure with the two optimizers *)

fun testfl_ (h :'aor hierarchy) ((pos,neg) :posneg) :posneg list =
	let val phi_h = build (InclAND(#1 h))
	    val phi_f = build_flat (InclAND(#1 h))
	    val flat  = posneg_closure h (pos,neg)
	in optimize_conds (phi_f,phi_h) flat end

fun testfl1_ (h :'aor hierarchy) ((pos,neg) :posneg) :posneg list =
	let val a     = #1 (fl_hierarchy h)
	    val sat   = build_tauth h
	    val flat  = posneg_closure h (pos,neg)
	in optimize1_conds (a,sat) flat end





local (* main function flattening transitions *)

(* CAUTION: some of the folds may seem inoptimal, but their direction      *)
(* has been carefully chosen to match the semantics of Interpr and         *)
(* C1, so that the same test cases can be used (mostly). Some of the folds *)
(* can probably be reverted, decreasing the complexity, but these are not  *)
(* real bottle necks of this algorithm, and correspondance with CodGenC1   *)
(* would be even harder to establish.                                      *)

  
  (* internal type of nearly flat transitions *)

  type ft = {action :action, refpos: stateid list, targets: stateid list}
  
  (* following 3 functions built transitions for default entry *)		    
                                                                                
  fun mkenor (h :attr hierarchy) (denv :denv) (ORState s,(tr,trs)) 
							:ft * ft list =
      let val {refpos,action,targets} = tr
          val def = getandh h (SE.lookup denv (#name s))                                 
          val t1 = {refpos=refpos,
		    action=appendact (enand def) action,                                 
		    targets=nameand def::targets}                                        
      in                                                                                 
		if deephor (ORState s) andalso enpureor (ORState s)                        
		then (tr,trs) 
		else if ishistoryor (ORState s) then                                       
			(tr,foldr (mkhis h denv tr) trs (#children s))                              
		else mkenand h denv (def,(t1,trs))                                                
      end  handle e => intErr e 200307113                                                

  and mkenand (h :attr hierarchy) (denv :denv) (ANDState s,(tr,trs)) 
							:ft * ft list =
	foldr (mkenor h denv) (tr,trs) (#children s)
	handle e => intErr e 200307142                                             

  and mkhis (h :attr hierarchy) (denv :denv) (tr :ft) (ANDState s,trs) 
							:ft list =                                                             
    let val t1 = { refpos= #name s:: #refpos tr, targets= [], 
                   action= #entry s }
        val (t2,trs2) = mkenand h denv (ANDState s,(t1,trs))                                             
    in if #action t2 = empty_action then trs2 else t2::trs2 end
    handle e => intErr e 200307112              

  (* following 2 functions built transitions for entry guided
     with explicit transition's targets. As soon as the guidence
     is over they hand over computation to above unguided functions*)
	     
  fun mkenor_guided (h :attr hierarchy) (denv :denv) (s,(ts,tr,trs)) 
					:stateid list * ft * ft list =
      let val (ts1,ts2) = List.partition (ancestorhst h (nameor s)) ts
          val {refpos,action,targets} = tr
          (* find is faster but filter to detect bugs *)
          val act = List.filter (fn k=>ancestorhst h (nameand k) 
		(hd ts1)) (childrenor s) handle Empty=>[]
      in case act of
	  [] => let val (tr1,trs1) = mkenor h denv (s,(tr,trs))
		in (ts2,tr1,trs1) end
	| [k] => let val tr1 = {refpos=refpos,
				action= appendact (enand k) action,
				targets= nameand k::targets}
		    val (_,tr2,trs2) = 
				mkenand_guided h denv (k,(ts1,tr1,trs))
		    in (ts2,tr2,trs2) end
	| ks => impossible "activate multiple children of ORst?"
      end handle e => intErr e 200307111

  and mkenand_guided (h :attr hierarchy) (denv :denv) (s,(ts,tr,trs))
		:stateid list * ft * ft list = 
	foldr (mkenor_guided h denv) (ts,tr,trs) (childrenand s)
	handle e => intErr e 200307143

                          

  (* flatten exits *)							    
                                                                      
  fun mkexor_guided (h :attr hierarchy) (ORState s, (ss,tr,trs)) 
				:stateid list * ft * ft list =
      let val (ss1,ss2) = List.partition (ancestorhst h (#name s)) ss          
          val {refpos,action,targets} = tr                      
          (* find is faster but filter to detect bugs *)                       
          val act = List.filter (fn k=>ancestorhst h (nameand k)               
		(hd ss1)) (#children s) handle Empty => []               
      in case act of                                                           
	  [] => (ss2,tr,mkexor tr (ORState s,trs))                          
	| [ANDState k] =>                                                
		let val tr1 = {refpos=refpos, targets=targets,
			       action =appendact action (#exit k)}
		    val retval = mkexand_guided h (ANDState k,(ss1,tr1,trs))
		in (ss2,#2 retval, #3 retval) end                        
	| ks => impossible "#200307232" (* multiple active children of ORst! *)                  
      end handle e => intErr e 200307114                                       

  and mkexand_guided (h :attr hierarchy) (ANDState s, (ss,tr,trs)) 
				:stateid list * ft * ft list =
      foldl (mkexor_guided h) (ss,tr,trs) (#children s)                            
      handle e => intErr e 200307116                                           

  and mkexor (tr :ft) (s,e) :ft list =                                              
      (if expureor s then e else foldl (mkexand tr) e (childrenor s))               
      handle e => intErr e 200307140                                           

  and mkexand (tr :ft) (ANDState {exit,children,name,...},e) :ft list =            
      let val t1 = {refpos= name:: #refpos tr, action= exit, targets=[]}
          val e1 = if exit = empty_action then e else t1::e 
      in foldl (mkexor t1) e1 children end handle e => intErr e 200307141           


in 

   fun fl_tran (h :attr hierarchy) (denv :denv) (a :F.anatomy, sat:mconf) 
	(t :static tran, cont :F.tran list)  :F.tran list =
	let 
	    val {trigger,pos,neg,guard,action,targets,annotation,scope}=t
		  
	    val tgts = map (fn st=>(#tgts st,getorh h (#stscope st),
				    valOf (#stsrcs st))) annotation
		       handle Option.Option => impossible "#200307240"
	    val trac :ft= {refpos=[], action=action, targets=[]}
	    val (_,tren,trens) = foldl (fn((ts,s,ss),(_,tr,trs))=>
			mkenor_guided h denv (s,(ts,tr,trs)))  ([],trac,[]) tgts
	    val (_,trex,trexs) = foldr (fn((_,s,ss),(_,tr,trs))=>
				(mkexor_guided h (s,(ss,tr,trs))))
				([],tren,[]) tgts 
	    val clo  = posneg_closure h (pos,neg)
	    val cloo = optimize1_conds (a,sat) clo
	    fun mktrans ({refpos,action,targets} :ft,cont) :F.tran list =  
		let fun mktran (p,n) =
			{trigger= trigger, guard= guard,
			 action=action, targets= targets, pos=p, neg=n}
		    val cloor  = map (fn (p,n)=> (refpos @p,neg)) cloo
		    val clooro = optimize1_conds (a,sat) cloor
		in map mktran clooro @ cont end

	(* CAUTION: in fact the transition list is generated reversed      *)
	(* this is because CodGenCF will reverse it once again on the way  *)
	in foldl mktrans cont (trexs @ trex::trens) end 
	handle e => intErr e 200307110
end

(* iterates transition's flattener over all transitions            *)


fun fl_trans (sys :(attr,static) system) (a :F.anatomy) :F.tran list =
	let val h     = #hierarchy sys 
	    val sat   = build_tauth h
	    val phi_h = build (InclAND (gettree h))
	    val trans = StaticScope.minsrcs_trans (phi_h, h) (#trans sys)
	in foldr (fl_tran h (#denv sys) (a,sat)) [] trans end

	
(* generate reset transition *)
(* cannot be done by transition flattener, bcs of different treatment of
   history states and or-states in general *)

fun fl_init (h :attr hierarchy) (denv :denv) : action * stateid list =
	let fun mkenor (ORState s,a) =
		mkenand (getandh h (SE.lookup denv (#name s)),a)
	    and mkenand (ANDState s,a) = 
		let val a1 = appendact (#entry s) a (* YUCK! *)
		in foldr mkenor a1 (#children s) end
	    fun fo (ORState s,ss) = SE.lookup denv (#name s)::ss
	    val r = gettree h
	in (mkenand (r,empty_action), hfoldland fo (fn (_,ss)=>ss) [] r) end
	handle e => intErr e 200307144


fun fl_sys (sysa : (attr,static) system) : F.system = 
	let val h            = #hierarchy sysa
	    val (a,machines) = fl_hierarchy h 
	in { 
		name    = #name sysa,
		fname   = #fname sysa,
		machines= machines,	 
		anatomy = a,        
		vartypes= #vartypes sysa, 
		evtypes = #evtypes sysa,  
		evgroups= #evgroups sysa, 
		otypes  = #otypes sysa,   
		denv    = #denv sysa,  (* some are needless *)   
		venv    = #venv sysa,     
		cenv    = #cenv sysa,
		init    = fl_init h (#denv sysa),
		trans   = fl_trans sysa a }
	end

fun flatten (model : (attr,static) model) :Flat.model = map fl_sys model
	


end
