(* 
   File   : /home/cvs/cvsroot/projects/visualstate/scope/parsehsm.sml,v
   Comment: A parser for HSM language
   Release: scope-0-16_final 

     Copyright (C) 2003 Andrzej Wasowski
     
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
     
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
     
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

     USAGE: to parse any file with this parser

	mosml -P full  Parsercomb.sig Parsercomb.sml Hsm.sml parsehsm.sml
	>parseFile "filename";
 
     The result should be a syntax tree.

   parsehsm.sml,v 1.2 2003/05/07 13:06:44 wasowski Exp
*)

structure ParseHSM : PARSEHSM = struct

open Parsercomb Hsm 


infix 6 $-- --$ #-- --#
infix 5 --
infix 3 >> >>*
infix 2 >>=
infix 0 ||

(* helper functions *)

fun isAlphaNumUndsc c = (Char.isAlphaNum c) orelse (c = #"_")
fun isAlphaUndsc    c = (Char.isAlpha c   ) orelse (c = #"_")
fun notAsterisk c = c <> #"*"
fun notSlash c = c <> #"/"

(* scan comments and white space *)

fun endComment _ =
	getChars0 notAsterisk #-- ($ "*/" || getItem >>= endComment) 

val skipComment = 
	("/*" $-- endComment #"_") >> (fn _ => NONE)

fun focus phrase _ = skipWS (phrase || skipComment >>= focus phrase)

(* scan identifiers, keywords and operators/delimiters *)

val scan_id = focus (getElem isAlphaUndsc -- getElems0 isAlphaNumUndsc 
						>> (implode o op::)) NONE

val EVENT    = focus ($ "@event" )   NONE
val ACTION   = focus ($ "@action")   NONE 
val INITIAL  = focus ($ "@initial")  NONE >> (fn _ => true)
val ANDSTATE = focus ($ "@andstate") NONE >> (fn _ => false)
val ORSTATE  = focus ($ "@orstate")  NONE
val LBR      = focus ($ "{")         NONE
val RBR      = focus ($ "}")         NONE
val SLASH    = focus ($ "/")         NONE
val COLON    = focus ($ ":")         NONE
val DOT      = focus ($ ".")         NONE
val ARROW    = focus ($ "->")        NONE

(* qualified state names *)

val spath   = scan_id -- repeat0 (DOT #-- scan_id) >> op:: 

(* declarations section *)

val events  = EVENT  #-- (repeat1 scan_id >> op::)
val actions = ACTION #-- (repeat1 scan_id >> op::)

val decl    = events >> (fn e=> (e,[])) || actions >> (fn a=> ([],a))

val decls   = repeat0 decl >> (fn l=> let val (e,a) = ListPair.unzip l
				      in (List.concat e,List.concat a) end)

(* transitions *)

val actions = optional (SLASH #-- repeat0 scan_id) 
		>> (fn acs => case acs of SOME a => a | _ => [])

val tran    = scan_id -- actions --# COLON -- spath --#
              ARROW -- spath >>
		(fn (((e,acs),src),tgt)=> {event=e, action=acs,
					   source=src, target=tgt})
(* state hierarchy *)

fun andstate () = ANDSTATE -- scan_id >>= andbody 
	       || ANDSTATE -- scan_id >> (fn (ini,s)=>AND(s,ini,[]))
	       || INITIAL  -- scan_id >>= andbody
	       || INITIAL  -- scan_id >> (fn (ini,s)=>AND(s,ini,[]))

and orbody s = (repeat1 (andstate()) >> op::) -- repeat0 tran
		>> (fn (ands,trans) => OR(s,ands,trans))

and orstate  () = (ORSTATE #-- scan_id --# LBR >>= orbody) --# RBR

and andbody (ini,s) = LBR #-- repeat0 (orstate()) --# RBR 
			>> (fn ors => AND (s,ini,ors))

val sem     = INITIAL -- scan_id >>= andbody 

(* start symbol *)

val hsm     = (decls -- sem) --# focus (eof()) NONE
		>> (fn ((e,a),s)=> {events=e,actions=a,hierarchy=s})

(* external interface *)

fun parseFile fname :hsm option = 
	let val h = TextIO.openIn fname
	    val s = TextIO.inputAll h
	    val _ = TextIO.closeIn h
	in scanString hsm s end

end
